"use strict";

// Read config.js
const config = require("./config");

const { graphql, buildSchema } = require("graphql");

// Express HTTP Server
const express = require("express");
const graphqlHTTP = require("express-graphql");

const mysql = require("mysql");
const conn = mysql.createConnection(config.dbsettings);
 
conn.connect();


// GraphQL schema
// Describes all types and field names
// Fields on Query and Mutation are root fields
const schema = buildSchema(`
  type Query {
    movie(id: ID!): Movie
	movieList(num: Int = 10, offset: Int = 0, order: String = "title", search: String, genres: [String]): MovieList!
  }
  
  type Mutation {
	  postReview(review: ReviewInput!): Boolean!
  }
  
  type Movie {
	  id: ID!
	  title: String!
	  origTitle: String
	  isAdult: Boolean!
	  year: Int
	  runtime: Int
	  crew: [Person]
	  reviews: [Review]
	  genres: [String]
	  avgRating: Float
	  numVotes: Int!
  }
  
  type MovieList {
	  movies: [Movie]
	  rowcount: Int!
  }
  
  type Review {
	  id: ID!
	  forMovie: ID!
	  reviewerName: String
	  text: String
	  score: Int
  }
  
  input ReviewInput {
	  forMovie: ID!
	  reviewerName: String
	  text: String
	  score: Int
  }
  
  type Person {
	  name: String!
	  starsIn: [Movie]
	  workedOn: [Movie]
	  role: String!
  }
`);


// Executes SQL query and returns a Promise that resolves with the result.
function fetchSQL(sql) {
	console.log("fetching", sql);
	
	return new Promise((resolve, reject) => {
		conn.query(sql, (error, results, fields) => {
			if (error) {
				reject(error);
			} else {
				resolve(results);
			}
		});
	});
}

// The root resolver contains all root fields,
// both query fields AND mutation fields
class RootResolver {
	/*** QUERIES ***/
	/***************/
	movie(args) {
		const query = mysql.format("SELECT id, title, origTitle, isAdult, year, runtime, genres, avgRating, numVotes FROM movies WHERE id = ?;", [args.id]);

		return fetchSQL(query).then((results) => {
			return new Movie(results[0]);
		});
	}
	
	// Returns a list of movies
	movieList(args) {
		let num = args.num;
		if (num < 1 || num > 10) {
			num = 10;
		}
		
		let order;
		let where = "";
		let offset = "OFFSET ?";
		
		let variables = [];
		
		if (args.order === "random") {
			order = "RAND()";
			offset = ""; // offset makes no sense for random results
		} else if (args.order === "score") {
			order = "avgRating DESC, numVotes DESC, title ASC";
		} else {
			order = "title ASC";
		}
		
		if (args.search || args.genres) {
			if (args.search) {
				where += " AND title LIKE ?";
				variables.push(`%${args.search}%`);
			}
			
			if (args.genres) {
				for (const genre of args.genres) {
					where += " AND FIND_IN_SET(?, genres) > 0";
					variables.push(genre);
				}
			}
		}

		variables.push(num);
		variables.push(args.offset);

		const countQuery = mysql.format(`SELECT COUNT(*) AS rowcount FROM movies WHERE NOT isAdult AND year IS NOT NULL ${where};`, variables);

		const query = mysql.format(`SELECT id, title, origTitle, isAdult, year, runtime FROM movies WHERE NOT isAdult AND year IS NOT NULL ${where} ORDER BY ${order} LIMIT ? ${offset};`, variables);


		// Lazy evaluation
		// movies and rowcount are functions that are only
		// called when these fields are requested
		return {
			movies: () => fetchSQL(query).then((results) => {
				return results.map(r => new Movie(r));
			}),
			rowcount: () => fetchSQL(countQuery).then((results) => {
				return results[0].rowcount;
			})
		};
	}
	

	/*** MUTATIONS ***/
	/*****************/
	postReview(args) {
		const review = args.review;
		const query = mysql.format("INSERT INTO reviews (reviewerName, reviewText, score, idmovie) VALUES (?, ?, ?, ?);", [review.reviewerName, review.text, review.score, review.forMovie]);
		return fetchSQL(query).then(r => true).catch(r => false);
	}
}


// Superclass for all objects that are populated from the database.
class DatabaseObject {
	// Basic constructor that copies all fields from a given object.
	// Useful for quickly converting a database query result into a GraphQL object.
	constructor(data) {
		for (const key of Object.keys(data)) {
			this[key] = data[key];
		}
	}
}


class Movie extends DatabaseObject {
	constructor(data) {
		super(data);
		
		// MySQL returns genres as comma-delimited string,
		// so turn it into an array
		if (this.genres) {
			this.genres = this.genres.split(",");
		}
	}
	
	// Get people who worked on this movie
	crew() {
		const query = mysql.format("SELECT p.id, p.name, c.role, c.job FROM crew c INNER JOIN people p on p.id = c.idperson INNER JOIN movies m on m.id = c.idmovie WHERE m.id = ?;", [this.id]);
	
		return fetchSQL(query).then((results) => {
			return results.map(r => new Person(r));
		});
	}
	
	// Get reviews for this movie
	reviews() {
		const query = mysql.format("SELECT reviewerName, reviewText AS text, score FROM reviews WHERE idmovie = ?;", [this.id]);
		
		return fetchSQL(query).then((results) => {
			return results.map(r => new Review(r));
		});
	}
}


class Review extends DatabaseObject {}


class Person extends DatabaseObject {
	// Get movies this person worked on
	workedOn() {
		const query = mysql.format("SELECT m.id, title, origTitle, isAdult, year, runtime FROM movies m INNER JOIN crew c ON c.idmovie = m.id WHERE c.idperson = ?;", [this.id]);
		
		return fetchSQL(query).then((results) => {
			return results.map(r => new Movie(r));
		});
	}
}



const webapp = express();

webapp.use("/graphql", graphqlHTTP({
	schema: schema,
	rootValue: new RootResolver(),
	context: { conn },
	graphiql: true,
}));

webapp.use(express.static("static"));

webapp.listen(8084);
