(function () {
	"use strict";
	
	function graphql(query, variables) {
		const body = { query };
		
		if (variables) {
			body.variables = variables;
		}
		
		return fetch("/graphql", {
			method: "POST",
			headers: {
				"Content-Type": "application/json"
			},
			body: JSON.stringify(body),
			mode: "same-origin"
		}).then((result) => result.json());
	}


	// Fetches poster image from OMDb
	const getPosterURL = (() => {
		const cache = {};
		
		return function (id) {
			// Convert numeric ID to IMDb ID
			// 503 -> tt00000503
			const imdbID = "tt" + String(id).padStart(8, "0");
			
			// Return URL immediately if we already fetched it previously
			if (cache[imdbID]) {
				return Promise.resolve(cache[imdbID]);
			}
			
			return fetch(`https://www.omdbapi.com/?apikey=bb80ca7&i=${imdbID}&type=movie`, {
				method: "GET",
				mode: "cors",
				referrer: "no-referrer"
			}).then(response => response.json())
			.then(response => {
				let finalURL = null;

				if (response.Poster && response.Poster.toLowerCase() !== "n/a") {
					finalURL = response.Poster;
				}
				
				return cache[imdbID] = finalURL;
			});
		};
	})();
	
	
	const movieList = document.getElementById("movielist");
	const detailview = document.getElementById("detailview");
	const moviedets = document.getElementById("moviedets");
	
	const searchPage = document.getElementById("page-search");
	const pageNumberContainer = document.getElementById("pagenumber");
	
	const navList = document.getElementById("navlist");
	let activeNav = document.getElementById("homenav");
	
	const detailTitle = document.getElementById("dt-title");
	const detailAka = document.getElementById("dt-aka");
	
	const searchForm = document.getElementById("searchform");
	const reviewForm = document.getElementById("reviewform");
	
	
	function makeMovieTable(data) {
		const table = document.createElement("table");
		
		table.border = "1";
		
		table.addEventListener("click", (event) => {
			const target = event.target.closest("tr");
			
			if ("movieid" in target.dataset) {
				showDetailView(target.dataset.movieid);
			}
		});
		
		table.className = "movietable";
		
		for (const movie of data.movieList.movies) {
			addMovieRow(table, movie);
		}
		
		return table;
	}
	
	
	function addMovieRow(table, movie) {
		const { title, origTitle, year, id } = movie;
		const directors = movie.crew.filter((person) => person.role === "director").map((person) => person.name).join("<br>");
		
		const row = table.insertRow();
		row.dataset.movieid = id;
		row.className = "clickable";
		
		const titleCell     = row.insertCell();
		const yearCell      = row.insertCell();
		const directorCell  = row.insertCell();
		
		titleCell.innerHTML = title;
		
		if (origTitle && origTitle !== title) {
			titleCell.innerHTML += `<br><small>${origTitle}</small>`;
		}
		
		yearCell.appendChild(document.createTextNode(year));
		directorCell.innerHTML = directors;
	}

	
	// Turns numerical scores into stars out of 5
	// 5 -> ★★★★★
	// 3 -> ★★★☆☆
	function scoreToStar(score) {
		return "★".repeat(score) + "☆".repeat(5 - score);
	}
	
	function escapeHTML(string) {
		const frag = document.createElement("p");
		frag.appendChild(document.createTextNode(string));
		
		return frag.innerHTML;
	}
	
	function makeMovieListing(movie) {
		return `<div class="movie" data-id="${movie.id}">
			<h2>${movie.title} (${movie.year})</h2>
		</div>`;
	}
	
	function makeDetailView(movie) {
		let aka = "";
		if (movie.title !== movie.origTitle) {
			aka = `<h3>(${movie.origTitle})</h3>`;
		}
		
		let rating = "";
		if (movie.avgRating) {
			rating = `<p>${movie.avgRating} Durchschnitt von ${movie.numVotes} Bewertungen</p>`;
		}
		
		let crew = "";
		for (const person of movie.crew) {
			crew += `<li>${person.role}: ${person.name}</li>`;
		}
		
		let reviews = "";
		
		if (movie.reviews && movie.reviews.length) {
			reviews += "<fieldset><legend>Bewertungen</legend>";
			for (const rev of movie.reviews) {
				reviews += `<p>${escapeHTML(rev.reviewerName || "Anonymous")} sagt:</p>`;
				reviews += `<p><strong>${scoreToStar(rev.score)}</strong> "${escapeHTML(rev.text)}"</p>`;
			}
			reviews += "</fieldset>";
		}
		
		
		return `<div class="movie-dt">
			<div class="poster" id="poster-${movie.id}"></div>
			<h2>${movie.title} (${movie.year})</h2>
			${aka}
			${rating}
			<ul>${crew}</ul>
			${reviews}
		</div>`;
	}
	
	
	function showDetailView(id) {
		graphql(`query($id: ID!) {
			movie(id: $id) {
				id title year origTitle genres avgRating numVotes
				reviews {
					reviewerName text score
				}
				crew {
					name role
					workedOn {
						title
					}
				} }
		}`, { id })
		.then(result => {
			moviedets.innerHTML = makeDetailView(result.data.movie);
			
			getPosterURL(id).then(url => {
				if (url) {
					const img = new Image();
					img.src = url;
					document.getElementById("poster-" + id).appendChild(img);
				}
			});
			
			// Update review form field
			reviewForm.elements.forMovie.value = id;
			
			switchPage("detailview");
		});
	}

	
	movieList.addEventListener("click", (event) => {
		const target = event.target.closest(".movie");
	
		
		if (target) {
			showDetailView(target.dataset.id);
		}
	});
	
	// Switch active page
	let activeSection = document.getElementById("page-home");
	
	function switchPage(page) {
		const newSection = document.getElementById(`page-${page}`);
		
		if (activeSection !== newSection) {
			activeSection.hidden = true;
			newSection.hidden    = false;
			
			activeSection = newSection;
		}
	}
	
	
	navList.addEventListener("click", (event) => {
		const target = event.target.closest("li");
		
		if (target && "page" in target.dataset) {
			activeNav.classList.remove("active");
			
			activeNav = target;
			activeNav.classList.add("active");
			
			switchPage(target.dataset.page);
		}
	});
	
	
	function paginationClick() {
		const target = event.target;
		
		if ("nav" in target.dataset) {
			const context = target.closest(".page");
			const params = context.dataset;

			let pageNumber = parseInt(params.page || 1, 10);
			
			pageNumber += (target.dataset.nav === "back" ? -1 : 1);
			
			if (pageNumber) {
				params.page = pageNumber;
				updateSearchView(context);
			}
		}
	}
	
	
	function updateSearchView(context) {
		const params = context.dataset;
		
		const pageNumberContainer = context.getElementsByClassName("pagenumber")[0];
		const resultsContainer    = context.getElementsByClassName("results")[0];
		
		pageNumberContainer.innerText = params.page || "1";
		
		searchMovies(params).then((response) => {
			processMovieListResponse(resultsContainer, response);
		});
	}
	
	for (const p of document.getElementsByClassName("pagination")) {
		p.addEventListener("click", paginationClick);
	}
	
	
	function searchMovies(params) {
		const MOVIES_PER_PAGE = 10;
		
		return graphql(`query($search: String, $num: Int, $offset: Int, $genres: [String], $order: String) {
			movieList(search: $search, num: $num, offset: $offset, genres: $genres, order: $order) {
				rowcount
				movies {
					id
					title
					year
					crew {
						name
						role
					}
				}
			}
		}`, {
			search: params.search || "",
			num: MOVIES_PER_PAGE,
			order: params.order || "title",
			genres: (params.genres ? [params.genres] : null),
			offset: (parseInt(params.page || 1, 10) - 1) * MOVIES_PER_PAGE
		});
	}
	
	
	searchForm.addEventListener("submit", (event) => {
		const target = event.target;
		const context = target.closest(".page");
		const params = context.dataset;
		
		// Change search parameters
		// Page 1 since this is a new search
		params.search = target.elements.search.value;
		params.page   = "1";
		
		updateSearchView(context);
		
		event.preventDefault();
	});
	
	reviewForm.addEventListener("submit", (event) => {
		const elements = event.target.elements;
		
		
		graphql(`mutation($review: ReviewInput!) {
			postReview(review: $review)
		}`, { review: { forMovie: elements.forMovie.value, reviewerName: elements.name.value, text: elements.text.value, score: parseInt(elements.score.value, 10) } });
		
	
		event.preventDefault();
	});
	
	
	
	function processMovieListResponse(targetElement, response) {
		targetElement.innerHTML = "";
		targetElement.appendChild(makeMovieTable(response.data));
	}
	
		
	// Fetch 5 random movies for the home page
	graphql(`{
		movieList(num: 5, order: "random") {
			rowcount
			movies {
				id
				title
				origTitle
				year
				crew {
					name
					role
				}
			}
		}
	}`).then((response) => {
		processMovieListResponse(movieList, response);
	});
	
	// Update all list views once. Could be done on-demand only or batched
	for (const section of ["top", "allmovies", "genre-horror", "genre-action", "genre-sport"]) {
		updateSearchView(document.getElementById("page-" + section));
	}
})();
